package de.tu_darmstadt.tk.SmartHomeNetworkSim.view;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.util.Collection;
import java.util.HashMap;
import java.util.Observable;
import java.util.Observer;

import javax.swing.JPanel;

import de.tu_darmstadt.tk.SmartHomeNetworkSim.control.ConfigurationController;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.control.Controller;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Connection;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Link;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Model;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Port;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.SmartDevice;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.util.Pair;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.view.VisualisationInteractor;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.view.util.Utility;

/**
 * Panel which visualizes the SmartHome Network, it's Devices and Connections
 * 
 *
 * @author Andreas T. Meyer-Berg
 */
@SuppressWarnings("serial")
public class VisualisationPanel extends JPanel implements Observer {

	
	
	/**
	 * Smart Home model which is visualized
	 */
	private Model model;

	/**
	 * Controller to notify when the model changes
	 */
	private Controller control;
	
	/**
	 * Configurations
	 */
	private ConfigurationController config;

	/**
	 * Listener which processes the GUI Interactions
	 */
	private VisualisationInteractor interactor;
	
	/**
	 * Initializes the Visualization Panel
	 * 
	 * @param model
	 *            Model to visualize
	 * @param control
	 *            Control, which changes the model
	 */
	public VisualisationPanel(Model model, Controller control) {
		super();

		this.model = model;
		this.control = control;
		this.config = control.getControllerConfiguration();
		
		this.interactor = new VisualisationInteractor(model, control, this);
		this.addMouseMotionListener(interactor);
		this.addMouseListener(interactor);
		this.addKeyListener(interactor);
	}

	/**
	 * Performs further initializations, which have to be performed, after the
	 * model was made visible
	 */
	public void delayedInit() {
		this.addComponentListener(new ComponentListener() {

			@Override
			public void componentShown(ComponentEvent e) {
			}

			@Override
			public void componentResized(ComponentEvent e) {
				control.setDimension(getWidth(), getHeight(), model.getDepth(),
						true);
				repaint();
			}

			@Override
			public void componentMoved(ComponentEvent e) {
			}

			@Override
			public void componentHidden(ComponentEvent e) {
			}
		});
		control.setDimension(getWidth(), getHeight(), model.getDepth(),
				true);
		repaint();
	}

	@Override
	public void paint(Graphics g) {
		// paint white background
		g.setColor(Color.white);
		g.fillRect(0, 0, this.getWidth(), this.getHeight());

		if(config.isShowConnections())
			paintConnections(g);
		if(config.isShowSmartDevices())
			paintDevices(g);
		
		paintDrag(g);
		
		g.setColor(Color.RED);
		String modus;
		switch (interactor.mode) {
		case VisualisationInteractor.DRAG_CONNECTION:
			modus="Drag Connection";
			break;
		case VisualisationInteractor.SELECTED_DRAG:
			modus="Drag Device";
			break;
		case VisualisationInteractor.DRAG_SELECT:
			modus="Drag Select";
			break;
		case VisualisationInteractor.NOTHING:
			modus="Nothing";
			break;
		case VisualisationInteractor.RIGHTCLICK_MENU:
			modus="RightClick Menu";
			break;
		case VisualisationInteractor.SELECTED:
			modus="Selected";
			break;
		default:
			modus="Unknown";
			break;
		}
		g.drawString("Debug(Modus:"+modus+")", 0, 10);
		if(interactor.controlDown)
			g.drawString("Control down", 0, 20);
	}

	/**
	 * Paints the smart devices of the Model
	 * 
	 * @param g
	 *            the Graphics object to visualize on
	 */
	public void paintDevices(Graphics g) {
		/*
		 * Visualization calculations for all devices
		 */
		/**
		 * Number of Links in the Model
		 */
		int numberOfLinks = control.getLinks().size();
		/**
		 * Angle per Link in "linkSlice degrees"
		 */
		double linkSlice = Math.min(360.0 / numberOfLinks,90.0);
		/**
		 * Map of links to Colors and position of the Color
		 */
		HashMap<Link, Pair<Integer,Color>> linkColors = null;
		/**
		 * Dimension in pixels of the link ring
		 */
		int linkDim = 6;
		
		if(config.isShowLinks()){
			linkColors = new HashMap<Link, Pair<Integer,Color>>();
			int i = 0;
			for(Link l: control.getLinks()){
				/**
				 * Distinct Color as String containing the hex values (128 static ones should be enough)
				 */
				String colorHex = Utility.indexcolors[i%Utility.indexcolors.length];
				/**
				 * Color of the given Hex String (Base16)
				 */
				Color color =  new Color(
						Integer.valueOf(colorHex.substring(1,3),16),
			            Integer.valueOf(colorHex.substring(3,5),16),
			            Integer.valueOf(colorHex.substring(5,7),16));
				/**
				 * Put position and Color into the HashMap
				 */
				linkColors.put(l, new Pair<Integer, Color>(i,color));
				i++;
			}
		}
		
		/**
		 * Paint Devices
		 */
		
		//Values for dragging of multiple Devices
		/**
		 * Pixel offset of dragged devices on the x axis
		 */
		int x_offset = 0;
		/**
		 * Pixel offset of dragged devices on the y axis
		 */
		int y_offset = 0;
		if(interactor.mode == VisualisationInteractor.SELECTED_DRAG && !interactor.selectedDevices.isEmpty()){
			x_offset = interactor.dragged_x-interactor.dragged.getX();
			y_offset = interactor.dragged_y-interactor.dragged.getY();
		}
		//Paint Devices
		for (SmartDevice s: model.getDevices()) {
			/**
			 * x Position of the device
			 */
			int x = s.getX();
			/**
			 * y Position of the device
			 */
			int y = s.getY();
			if(interactor.mode==VisualisationInteractor.SELECTED_DRAG && interactor.selectedDevices.contains(s)) {
				// Update visualization of dragged object
				x += x_offset;
				y += y_offset;
			}
			//Paint Links of Device V0.0
			if(config.isShowLinks()){
				for(Link l:s.getLinks()){
					/*
					 * Split circle around the object into NumberOfLinks Sections, color the sections, connected with the device
					 */
					/**
					 * Number and color of the current link
					 */
					Pair<Integer, Color> linkPos = linkColors.get(l);
					if(linkPos==null)continue;//Skip Links, which are not in the model
					g.setColor(linkPos.getRight());
					g.fillArc(x - config.getDeviceVisualizationRadius() - linkDim, y - config.getDeviceVisualizationRadius() - linkDim,
							2 * config.getDeviceVisualizationRadius() - 1 + 2 * linkDim, 2 * config.getDeviceVisualizationRadius() - 1 + 2*linkDim,
							(int)Math.round(linkPos.getLeft()*linkSlice), (int)Math.ceil(linkSlice));
				}
			
			}
			if(//(interactor.mode==VisualisationInteractor.SELECTED||interactor.mode==VisualisationInteractor.DRAG_SELECT ||interactor.mode == VisualisationInteractor.SELECTED_DRAG || interactor.mode == VisualisationInteractor.RIGHTCLICK_MENU)&&
					(interactor.selectedDevices.contains(s)^interactor.selectedDevicesDrag.contains(s))){
				//HighlightSelected Devices
				g.setColor(new Color(135,206,235));
			}else{
				g.setColor(Color.WHITE);
			}
			g.fillOval(x - config.getDeviceVisualizationRadius(), y - config.getDeviceVisualizationRadius(),
					2 * config.getDeviceVisualizationRadius() - 1, 2 * config.getDeviceVisualizationRadius() - 1);
			g.setColor(Color.BLACK);
			g.drawOval(x - config.getDeviceVisualizationRadius(), y - config.getDeviceVisualizationRadius(),
					2 * config.getDeviceVisualizationRadius() - 1, 2 * config.getDeviceVisualizationRadius() - 1);
			g.setColor(Color.BLUE);
			g.drawOval(x - config.getDeviceVisualizationRadius() + 2,
					y - config.getDeviceVisualizationRadius() + 2, 2 * config.getDeviceVisualizationRadius() - 5,
					2 * config.getDeviceVisualizationRadius() - 5);
			g.setColor(Color.BLACK);
			if(config.isShowSmartDeviceNames())
				g.drawString(s.getName(),
					x - g.getFontMetrics().stringWidth(s.getName()) / 2, y
							+ config.getDeviceVisualizationRadius() + 11 + (config.isShowLinks()?linkDim:0));
		}
	}
	
	/**
	 * Paints the Connections
	 * 
	 * @param g
	 *            the Graphics object to visualize on
	 */
	private void paintConnections(Graphics g) {
		//Values for dragging of multiple Devices
		int x_offset = 0;
		int y_offset = 0;
		if(interactor.mode == VisualisationInteractor.SELECTED_DRAG && !interactor.selectedDevices.isEmpty()){
			x_offset = interactor.dragged_x-interactor.dragged.getX();
			y_offset = interactor.dragged_y-interactor.dragged.getY();
		}
		
		// For all Connections
		for (Connection c : model.getConnections()) {
			Color connectionState;
			switch (c.getStatus()) {
			case Connection.ACTIVE:
				connectionState = Color.GREEN;
				break;
			case Connection.HALTED:
				connectionState = Color.ORANGE;
				break;
			case Connection.DONE:
				connectionState = Color.GRAY;
				break;
			case Connection.TERMINATED:
			case Connection.FINISHED:
				connectionState = Color.RED;
				break;
			default:
				connectionState = Color.BLUE;
				break;
			}
			g.setColor(connectionState);
			/**
			 * All Devices that are part of the connection
			 */
			Collection<Port> d = c.getParticipants();
			if(d.size()==0){
				System.out.println("WARNING: Empty, undeleted Connection: "+c.toString());
				continue;
			}
			if(c.getProtocol()==null){
				System.out.println("WARNING: Protocol is null: "+c.toString());
				continue;
			}
			for(Pair<Port,Port> p: c.getProtocol().getTopology() ){
				if(p.getLeft() != null && p.getLeft().getOwner() != null && p.getRight() != null && p.getRight().getOwner() != null){
					if(interactor.clickedConnection.contains(new Pair<Connection, Pair<Port,Port>>(c, p)))
						g.setColor(Color.BLUE);
					else
						g.setColor(connectionState);
					SmartDevice from = p.getLeft().getOwner();
					SmartDevice to = p.getRight().getOwner();
					int xFrom = from.getX();
					int yFrom = from.getY();
					int xTo = to.getX();
					int yTo = to.getY();
					if(interactor.selectedDevices.contains(from)^interactor.selectedDevicesDrag.contains(from)){
						xFrom+=x_offset;
						yFrom+=y_offset;
					}
					if(interactor.selectedDevices.contains(to)^interactor.selectedDevicesDrag.contains(to)){
						xTo+=x_offset;
						yTo+=y_offset;
					}
					g.drawLine(xFrom, yFrom, xTo, yTo);
				}
			}
			for(Pair<Port,Port> p: c.getProtocol().getDeletedTopology() ){
				if(interactor.clickedConnection.contains(new Pair<Connection, Pair<Port,Port>>(c, p)))
					g.setColor(Color.BLUE);
				else{
					//Just paint terminated connection, if config says so
					if(config.isShowTerminatedConnections())
						g.setColor(Color.RED);
					else
						continue;
				
				}
				if(p.getLeft() != null && p.getLeft().getOwner() != null && p.getRight() != null && p.getRight().getOwner() != null){
					SmartDevice from = p.getLeft().getOwner();
					SmartDevice to = p.getRight().getOwner();
					int xFrom = from.getX();
					int yFrom = from.getY();
					int xTo = to.getX();
					int yTo = to.getY();
					if(interactor.selectedDevices.contains(from)^interactor.selectedDevicesDrag.contains(from)){
						xFrom+=x_offset;
						yFrom+=y_offset;
					}
					if(interactor.selectedDevices.contains(to)^interactor.selectedDevicesDrag.contains(to)){
						xTo+=x_offset;
						yTo+=y_offset;
					}
					g.drawLine(xFrom, yFrom, xTo, yTo);
				}
			}
		}
		// paint new in progress connection, if a connection or link is in
		// creation
		if (interactor.mode == VisualisationInteractor.DRAG_CONNECTION && interactor.connectionFrom != null){
			g.setColor(Color.ORANGE);
			g.drawLine(interactor.connectionFrom.getX(),
					interactor.connectionFrom.getY(), interactor.dragged_x,
					interactor.dragged_y);
		}
	}
	
	/**
	 * Paints the 
	 * @param g
	 */
	public void paintDrag(Graphics g){
		if(interactor.mode!=VisualisationInteractor.DRAG_SELECT)
			return;	
		int low_x = Math.min(interactor.dragged_x, interactor.dragged_x_start);
		int low_y = Math.min(interactor.dragged_y, interactor.dragged_y_start);
		int high_x = Math.max(interactor.dragged_x, interactor.dragged_x_start);
		int high_y = Math.max(interactor.dragged_y, interactor.dragged_y_start);
		g.setColor(Color.BLUE);
		g.drawRect(low_x, low_y, high_x-low_x, high_y-low_y);

	}
	@Override
	public void update(Observable o, Object arg) {
		repaint();
	}

}
